@@ -18,22 +18,25 @@ with this program; if not, write to the Free Software Foundation, Inc.,
 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 -}
 
-{-# LANGUAGE FlexibleContexts, CPP, OverloadedStrings, RankNTypes #-}
+{-# LANGUAGE CPP, OverloadedLists #-}
 import Test.Hspec
 import Text.Pandoc hiding (getDataFileName)
-import Text.Pandoc.Builder
+import Text.Pandoc.Builder hiding (figure)
+import qualified Text.Pandoc.Builder as B
+import Control.Monad.Reader
 import Control.Monad.State
 import Data.List
 import Control.Arrow
 import qualified Data.Map as M
 import qualified Data.Text as T
-import qualified Data.Default as Df
+import Data.Maybe
 
 import Text.Pandoc.CrossRef
 import Text.Pandoc.CrossRef.Util.Options
 import Text.Pandoc.CrossRef.Util.Util
 import Text.Pandoc.CrossRef.References.Types
 import Lens.Micro
+import Text.Pandoc.CrossRef.References.Monad
 import qualified Text.Pandoc.CrossRef.References.Blocks as References.Blocks
 import qualified Text.Pandoc.CrossRef.References.Refs as References.Refs
 import qualified Text.Pandoc.CrossRef.References.List as References.List
@@ -50,8 +53,8 @@ main = hspec $ do
     describe "References.Blocks.replaceInlines" $ do
       it "Labels equations" $
         testAll (equation' "a^2+b^2=c^2" "equation")
-        (spanWith ("eq:equation", [], []) (equation' "a^2+b^2=c^2\\qquad(1)" ""),
-          eqnRefs =: M.fromList $ refRec'' "eq:equation" 1)
+        (spanWith ("eq:equation", [], []) (equation' "a^2+b^2=c^2\\qquad{(1)}" ""),
+          PfxEqn =: M.fromList $ refRec'' "eq:equation" 1)
       it "Labels equations in the middle of text" $
         testAll (
                 text "This is an equation: "
@@ -59,25 +62,25 @@ main = hspec $ do
              <> text " it should be labeled")
         (
            text "This is an equation: "
-        <> spanWith ("eq:equation", [], []) (equation' "a^2+b^2=c^2\\qquad(1)" "")
+        <> spanWith ("eq:equation", [], []) (equation' "a^2+b^2=c^2\\qquad{(1)}" "")
         <> text " it should be labeled",
-          eqnRefs =: M.fromList $ refRec'' "eq:equation" 1)
+          PfxEqn =: M.fromList $ refRec'' "eq:equation" 1)
       it "Labels equations in the beginning of text" $
         testAll (
                 equation' "a^2+b^2=c^2" "equation"
              <> text " it should be labeled")
         (
-           spanWith ("eq:equation", [], []) (equation' "a^2+b^2=c^2\\qquad(1)" "")
+           spanWith ("eq:equation", [], []) (equation' "a^2+b^2=c^2\\qquad{(1)}" "")
         <> text " it should be labeled",
-          eqnRefs =: M.fromList $ refRec'' "eq:equation" 1)
+          PfxEqn =: M.fromList $ refRec'' "eq:equation" 1)
       it "Labels equations in the end of text" $
         testAll (
                 text "This is an equation: "
              <> equation' "a^2+b^2=c^2" "equation")
         (
            text "This is an equation: "
-        <> spanWith ("eq:equation", [], []) (equation' "a^2+b^2=c^2\\qquad(1)" ""),
-          eqnRefs =: M.fromList $ refRec'' "eq:equation" 1)
+        <> spanWith ("eq:equation", [], []) (equation' "a^2+b^2=c^2\\qquad{(1)}" ""),
+          PfxEqn =: M.fromList $ refRec'' "eq:equation" 1)
 
     -- TODO:
     -- describe "References.Blocks.spanInlines"
@@ -85,63 +88,81 @@ main = hspec $ do
 
     describe "References.Blocks.replaceBlocks" $ do
       it "Labels images" $
-        testAll (figure "test.jpg" "" "Test figure" "figure")
-        (figure "test.jpg" "" "Figure 1: Test figure" "figure",
-          imgRefs =: M.fromList $ refRec' "fig:figure" 1 "Test figure")
+        testAll (figure "test.jpg" "" "Test figure" Nothing "figure")
+        (figure "test.jpg" "" "Figure 1: Test figure" (Just "Figure 1: Test figure") "figure",
+          PfxImg =: M.fromList $ refRec' "fig:figure" 1 "Test figure")
       it "Labels subfigures" $
         testAll (
           divWith ("fig:subfigure",[],[]) (
-            para (figure' "fig:" "test1.jpg" "" "Test figure 1" "figure1")
-          <>para (figure' "fig:" "test2.jpg" "" "Test figure 2" "figure2")
+            para (figure' "test1.jpg" "" "Test figure 1" "figure1")
+          <>para (figure' "test2.jpg" "" "Test figure 2" "figure2")
           <>para (text "figure caption")
             ) <>
           divWith ("fig:subfigure2",[],[]) (
-            para (figure' "fig:" "test21.jpg" "" "Test figure 21" "figure21")
-          <>para (figure' "fig:" "test22.jpg" "" "Test figure 22" "figure22")
+            para (figure' "test21.jpg" "" "Test figure 21" "figure21")
+          <>para (figure' "test22.jpg" "" "Test figure 22" "figure22")
           <>para (text "figure caption 2")
             )
           )
         (
-          divWith ("fig:subfigure",["subfigures"],[]) (
-               para (figure' "fig:" "test1.jpg" "" "a" "figure1")
-            <> para (figure' "fig:" "test2.jpg" "" "b" "figure2")
-            <> para (text "Figure 1: figure caption. a — Test figure 1, b — Test figure 2")
+          figureWith ("fig:subfigure",["subfigures"],[])
+            (caption Nothing $ para (text "Figure 1: figure caption. a — Test figure 1, b — Test figure 2"))
+            ( para (figure' "test1.jpg" "" "a" "figure1")
+            <> para (figure' "test2.jpg" "" "b" "figure2")
             ) <>
-          divWith ("fig:subfigure2",["subfigures"],[]) (
-               para (figure' "fig:" "test21.jpg" "" "a" "figure21")
-            <> para (figure' "fig:" "test22.jpg" "" "b" "figure22")
-            <> para (text "Figure 2: figure caption 2. a — Test figure 21, b — Test figure 22")
+          figureWith ("fig:subfigure2",["subfigures"],[])
+            (caption Nothing $ para (text "Figure 2: figure caption 2. a — Test figure 21, b — Test figure 22"))
+            (  para (figure' "test21.jpg" "" "a" "figure21")
+            <> para (figure' "test22.jpg" "" "b" "figure22")
             )
-        , imgRefs =: M.fromList [("fig:figure1",RefRec {
+        , PfxImg =: M.fromList [("fig:figure1",RefRec {
                                             refIndex = [(1,Nothing)],
+                                            refGlobal = 1,
                                             refTitle = [Str "Test",Space,Str "figure",Space,Str "1"],
+                                            refHideFromList = False,
+                                            refLabel = Just "fig:figure1",
                                             refSubfigure = Just [(1, Just "a")]}),
                                     ("fig:figure2",RefRec {
                                             refIndex = [(1,Nothing)],
+                                            refGlobal = 2,
                                             refTitle = [Str "Test",Space,Str "figure",Space,Str "2"],
+                                            refHideFromList = False,
+                                            refLabel = Just "fig:figure2",
                                             refSubfigure = Just [(2, Just "b")]}),
                                     ("fig:subfigure",RefRec {
                                             refIndex = [(1,Nothing)],
+                                            refGlobal = 0,
                                             refTitle = [Str "figure",Space,Str "caption"],
+                                            refHideFromList = False,
+                                            refLabel = Just "fig:subfigure",
                                             refSubfigure = Nothing}),
                                     ("fig:figure21",RefRec {
                                             refIndex = [(2,Nothing)],
+                                            refGlobal = 4,
                                             refTitle = [Str "Test",Space,Str "figure",Space,Str "21"],
+                                            refHideFromList = False,
+                                            refLabel = Just "fig:figure21",
                                             refSubfigure = Just [(1, Just "a")]}),
                                     ("fig:figure22",RefRec {
                                             refIndex = [(2,Nothing)],
+                                            refGlobal = 5,
                                             refTitle = [Str "Test",Space,Str "figure",Space,Str "22"],
+                                            refHideFromList = False,
+                                            refLabel = Just "fig:figure22",
                                             refSubfigure = Just [(2, Just "b")]}),
                                     ("fig:subfigure2",RefRec {
                                             refIndex = [(2,Nothing)],
+                                            refGlobal = 3,
                                             refTitle = [Str "figure",Space,Str "caption",Space,Str "2"],
+                                            refHideFromList = False,
+                                            refLabel = Just "fig:subfigure2",
                                             refSubfigure = Nothing})
                                    ]
             )
       it "Labels equations" $
         testAll (equation "a^2+b^2=c^2" "equation")
-        (para $ spanWith ("eq:equation", [], []) (equation' "a^2+b^2=c^2\\qquad(1)" ""),
-          eqnRefs =: M.fromList $ refRec'' "eq:equation" 1)
+        (para $ spanWith ("eq:equation", [], []) (equation' "a^2+b^2=c^2\\qquad{(1)}" ""),
+          PfxEqn =: M.fromList $ refRec'' "eq:equation" 1)
       it "Labels equations in the middle of text" $
         testAll (para $
                 text "This is an equation: "
@@ -149,98 +170,99 @@ main = hspec $ do
              <> text " it should be labeled")
         (para $
            text "This is an equation: "
-        <> spanWith ("eq:equation", [], []) (equation' "a^2+b^2=c^2\\qquad(1)" "")
+        <> spanWith ("eq:equation", [], []) (equation' "a^2+b^2=c^2\\qquad{(1)}" "")
         <> text " it should be labeled",
-          eqnRefs =: M.fromList $ refRec'' "eq:equation" 1)
+          PfxEqn =: M.fromList $ refRec'' "eq:equation" 1)
       it "Labels equations in the beginning of text" $
         testAll (para $
                 equation' "a^2+b^2=c^2" "equation"
              <> text " it should be labeled")
         (para $
-           spanWith ("eq:equation", [], []) (equation' "a^2+b^2=c^2\\qquad(1)" "")
+           spanWith ("eq:equation", [], []) (equation' "a^2+b^2=c^2\\qquad{(1)}" "")
         <> text " it should be labeled",
-          eqnRefs =: M.fromList $ refRec'' "eq:equation" 1)
+          PfxEqn =: M.fromList $ refRec'' "eq:equation" 1)
       it "Labels equations in the end of text" $
         testAll (para $
                 text "This is an equation: "
              <> equation' "a^2+b^2=c^2" "equation")
         (para $
            text "This is an equation: "
-        <> spanWith ("eq:equation", [], []) (equation' "a^2+b^2=c^2\\qquad(1)" ""),
-          eqnRefs =: M.fromList $ refRec'' "eq:equation" 1)
+        <> spanWith ("eq:equation", [], []) (equation' "a^2+b^2=c^2\\qquad{(1)}" ""),
+          PfxEqn =: M.fromList $ refRec'' "eq:equation" 1)
       it "Labels tables" $
         testAll (table' "Test table" "table")
         (divWith ("tbl:table", [], []) $ table' "Table 1: Test table" "",
-          tblRefs =: M.fromList $ refRec' "tbl:table" 1 "Test table")
+          PfxTbl =: M.fromList $ refRec' "tbl:table" 1 "Test table")
       it "Labels code blocks" $
         testAll (codeBlock' "Test code block" "codeblock")
         (codeBlockDiv "Listing 1: Test code block" "codeblock",
-          lstRefs =: M.fromList $ refRec' "lst:codeblock" 1 "Test code block")
+          PfxLst =: M.fromList $ refRec' "lst:codeblock" 1 "Test code block")
       it "Labels code block divs" $
         testAll (codeBlockDiv "Test code block" "codeblock")
         (codeBlockDiv "Listing 1: Test code block" "codeblock",
-          lstRefs =: M.fromList $ refRec' "lst:codeblock" 1 "Test code block")
+          PfxLst =: M.fromList $ refRec' "lst:codeblock" 1 "Test code block")
       it "Labels sections divs" $
         testAll (section "Section Header" 1 "section")
         (section "Section Header" 1 "section",
-          secRefs .~ M.fromList (refRec' "sec:section" 1 "Section Header")
-          $ curChap =: [(1,Nothing)])
+          PfxSec =: M.fromList (refRec' "sec:section" 1 "Section Header"))
 
     describe "References.Refs.replaceRefs" $ do
       it "References one image" $
-        testRefs' "fig:" [1] [4] imgRefs "fig.\160\&4"
+        testRefs' "fig:" [1] [4] PfxImg "fig.\160\&4"
       it "References multiple images" $
-        testRefs' "fig:" [1..3] [4..6] imgRefs "figs.\160\&4-6"
+        testRefs' "fig:" [1..3] [4..6] PfxImg "figs.\160\&4-6"
       it "References one equation" $
-        testRefs' "eq:" [1] [4] eqnRefs "eq.\160\&4"
+        testRefs' "eq:" [1] [4] PfxEqn "eq.\160\&4"
       it "References multiple equations" $
-        testRefs' "eq:" [1..3] [4..6] eqnRefs "eqns.\160\&4-6"
+        testRefs' "eq:" [1..3] [4..6] PfxEqn "eqns.\160\&4-6"
       it "References one table" $
-        testRefs' "tbl:" [1] [4] tblRefs "tbl.\160\&4"
+        testRefs' "tbl:" [1] [4] PfxTbl "tbl.\160\&4"
       it "References multiple tables" $
-        testRefs' "tbl:" [1..3] [4..6] tblRefs "tbls.\160\&4-6"
+        testRefs' "tbl:" [1..3] [4..6] PfxTbl "tbls.\160\&4-6"
       it "References one listing" $
-        testRefs' "lst:" [1] [4] lstRefs "lst.\160\&4"
+        testRefs' "lst:" [1] [4] PfxLst "lst.\160\&4"
       it "References multiple listings" $
-        testRefs' "lst:" [1..3] [4..6] lstRefs "lsts.\160\&4-6"
+        testRefs' "lst:" [1..3] [4..6] PfxLst "lsts.\160\&4-6"
       it "References one section" $
-        testRefs' "sec:" [1] [4] secRefs "sec.\160\&4"
+        testRefs' "sec:" [1] [4] PfxSec "sec.\160\&4"
       it "References multiple sections" $
-        testRefs' "sec:" [1..3] [4..6] secRefs "secs.\160\&4-6"
+        testRefs' "sec:" [1..3] [4..6] PfxSec "secs.\160\&4-6"
       it "Separates references to different chapter items by a comma" $
-        testRefs'' "lst:" [1..6] (zip [1,1..] [4..6] <> zip [2,2..] [7..9]) lstRefs "lsts.\160\&1.4-1.6, 2.7-2.9"
+        testRefs'' "lst:" [1..6] (zip [1,1..] [4..6] <> zip [2,2..] [7..9]) PfxLst "lsts.\160\&1.4-1.6, 2.7-2.9"
 
     describe "References.Refs.replaceRefs capitalization" $ do
       it "References one image" $
-        testRefs' "Fig:" [1] [4] imgRefs "Fig.\160\&4"
+        testRefs' "Fig:" [1] [4] PfxImg "Fig.\160\&4"
       it "References multiple images" $
-        testRefs' "Fig:" [1..3] [4..6] imgRefs "Figs.\160\&4-6"
+        testRefs' "Fig:" [1..3] [4..6] PfxImg "Figs.\160\&4-6"
       it "References one equation" $
-        testRefs' "Eq:" [1] [4] eqnRefs "Eq.\160\&4"
+        testRefs' "Eq:" [1] [4] PfxEqn "Eq.\160\&4"
       it "References multiple equations" $
-        testRefs' "Eq:" [1..3] [4..6] eqnRefs "Eqns.\160\&4-6"
+        testRefs' "Eq:" [1..3] [4..6] PfxEqn "Eqns.\160\&4-6"
       it "References one table" $
-        testRefs' "Tbl:" [1] [4] tblRefs "Tbl.\160\&4"
+        testRefs' "Tbl:" [1] [4] PfxTbl "Tbl.\160\&4"
       it "References multiple tables" $
-        testRefs' "Tbl:" [1..3] [4..6] tblRefs "Tbls.\160\&4-6"
+        testRefs' "Tbl:" [1..3] [4..6] PfxTbl "Tbls.\160\&4-6"
       it "References one listing" $
-        testRefs' "Lst:" [1] [4] lstRefs "Lst.\160\&4"
+        testRefs' "Lst:" [1] [4] PfxLst "Lst.\160\&4"
       it "References multiple listings" $
-        testRefs' "Lst:" [1..3] [4..6] lstRefs "Lsts.\160\&4-6"
+        testRefs' "Lst:" [1..3] [4..6] PfxLst "Lsts.\160\&4-6"
       it "References one listing" $
-        testRefs' "Sec:" [1] [4] secRefs "Sec.\160\&4"
+        testRefs' "Sec:" [1] [4] PfxSec "Sec.\160\&4"
       it "References multiple listings" $
-        testRefs' "Sec:" [1..3] [4..6] secRefs "Secs.\160\&4-6"
+        testRefs' "Sec:" [1..3] [4..6] PfxSec "Secs.\160\&4-6"
 
     describe "References.List.listOf" $ do
       it "Generates list of tables" $
         testList (rawBlock "latex" "\\listoftables")
-                 (tblRefs =: M.fromList $ refRec' "tbl:1" 4 "4" <> refRec' "tbl:2" 5 "5" <> refRec' "tbl:3" 6 "6")
-                 (header 1 (text "List of Tables") <> orderedList ((plain . str . T.pack . show) `map` [4..6 :: Int]))
+                 (PfxTbl =: M.fromList $ refRec' "tbl:1" 4 "4" <> refRec' "tbl:2" 5 "5" <> refRec' "tbl:3" 6 "6")
+                 (header 1 (text "List of Tables") <> divWith ("", ["list", "list-of-tbl"], [])
+                   (mconcat $ map (\n -> plain (str (T.pack $ show n <> ".") <> space <> str (T.pack $ show n) <> linebreak)) [4..6 :: Int]))
       it "Generates list of figures" $
         testList (rawBlock "latex" "\\listoffigures")
-                 (imgRefs =: M.fromList $ refRec' "fig:1" 4 "4" <> refRec' "fig:2" 5 "5" <> refRec' "fig:3" 6 "6")
-                 (header 1 (text "List of Figures") <> orderedList ((plain . str . T.pack . show) `map` [4..6 :: Int]))
+                 (PfxImg =: M.fromList $ refRec' "fig:1" 4 "4" <> refRec' "fig:2" 5 "5" <> refRec' "fig:3" 6 "6")
+                 (header 1 (text "List of Figures") <> divWith ("", ["list", "list-of-fig"], [])
+                   (mconcat $ map (\n -> plain (str (T.pack $ show n <> ".") <> space <> str (T.pack $ show n) <> linebreak)) [4..6 :: Int]))
 
     describe "Util.CodeBlockCaptions" $
       it "Transforms table-style codeBlock captions to codeblock divs" $ do
@@ -252,7 +274,7 @@ main = hspec $ do
 
     describe "Util.Template" $
       it "Applies templates" $
-        let template=Util.Template.makeTemplate defaultMeta (toList $ displayMath "figureTitle" <> displayMath "i" <> displayMath "t")
+        let (template :: Util.Template.Template)=Util.Template.makeTemplate defaultMeta (toList $ displayMath "figureTitle" <> displayMath "i" <> displayMath "t")
         in Util.Template.applyTemplate [Str "1"] [Str "title"] template `shouldBe`
            toList (str "Figure" <> str "1" <> str "title")
 
@@ -297,37 +319,48 @@ main = hspec $ do
         it "Section labels" $
           headerWith ("sec:section_label1", [], []) 1 (text "Section")
             <> para (citeGen "sec:section_label" [1])
-            `test` "\\hypertarget{sec:section_label1}{%\n\\section{Section}\\label{sec:section_label1}}\n\nsec.~\\ref{sec:section_label1}"
+            `test` "\\section{Section}\\label{sec:section_label1}\n\nsec.~\\ref{sec:section_label1}"
 
         it "Image labels" $
-          figure "img.png" "" "Title" "figure_label1"
+          figure "img.png" "" "Title" Nothing "figure_label1"
             <> para (citeGen "fig:figure_label" [1])
-            `test` "\\begin{figure}\n\\hypertarget{fig:figure_label1}{%\n\\centering\n\\includegraphics{img.png}\n\\caption{Title}\\label{fig:figure_label1}\n}\n\\end{figure}\n\nfig.~\\ref{fig:figure_label1}"
+            `test` "\\begin{figure}\n\\centering\n\\pandocbounded{\\includegraphics[keepaspectratio]{img.png}}\n\\caption{Title}\\label{fig:figure_label1}\n\\end{figure}\n\nfig.~\\ref{fig:figure_label1}"
 
+#ifdef FLAKY
         it "Eqn labels" $
           equation "x^2" "some_equation1"
             <> para (citeGen "eq:some_equation" [1])
-            `test` "\\begin{equation}\\protect\\hypertarget{eq:some_equation1}{}{x^2}\\label{eq:some_equation1}\\end{equation}\n\neq.~\\ref{eq:some_equation1}"
+            `test` "\\begin{equation}\\protect\\phantomsection\\label{eq:some_equation1}{x^2}\\end{equation}\n\neq.~\\ref{eq:some_equation1}"
 
-#ifdef FLAKY
         it "Tbl labels" $
           table' "A table" "some_table1"
             <> para (citeGen "tbl:some_table" [1])
-            `test` "\\hypertarget{tbl:some_table1}{}\n\\begin{longtable}[]{@{}@{}}\n\\caption{\\label{tbl:some_table1}A table}\\tabularnewline\n\\toprule\n\\endhead\n \\\\\n\\bottomrule\n\\end{longtable}\n\ntbl.~\\ref{tbl:some_table1}"
+            `test` concat (
+              [ "\\begin{longtable}[]{@{}l@{}}\n"
+              , "\\caption{\\label{tbl:some_table1}A table}\\tabularnewline\n"
+              , "\\toprule\\noalign{}\n"
+              , "\\endfirsthead\n"
+              , "\\endhead\n"
+              , "\\bottomrule\\noalign{}\n"
+              , "\\endlastfoot\n"
+              , " \\\\\n"
+              , "\\end{longtable}\n\n"
+              , "tbl.~\\ref{tbl:some_table1}"
+              ] :: [String])
 #endif
 
         it "Code block labels" $ do
           codeBlock' "A code block" "some_codeblock1"
             <> para (citeGen "lst:some_codeblock" [1])
-            `test` "\\begin{codelisting}\n\n\\caption{A code block}\n\n\\hypertarget{lst:some_codeblock1}{%\n\\label{lst:some_codeblock1}}%\n\\begin{Shaded}\n\\begin{Highlighting}[]\n\\OtherTok{main ::} \\DataTypeTok{IO}\\NormalTok{ ()}\n\\end{Highlighting}\n\\end{Shaded}\n\n\\end{codelisting}\n\nlst.~\\ref{lst:some_codeblock1}"
+            `test` "\\begin{codelisting}\n\n\\caption{A code block}\\label{lst:some_codeblock1}\n\n\\begin{Shaded}\n\\begin{Highlighting}[]\n\\OtherTok{main ::} \\DataTypeTok{IO}\\NormalTok{ ()}\n\\end{Highlighting}\n\\end{Shaded}\n\n\\end{codelisting}\n\nlst.~\\ref{lst:some_codeblock1}"
           codeBlock' "A code block with under_score" "some_codeblock1"
             <> para (citeGen "lst:some_codeblock" [1])
-            `test` "\\begin{codelisting}\n\n\\caption{A code block with under\\_score}\n\n\\hypertarget{lst:some_codeblock1}{%\n\\label{lst:some_codeblock1}}%\n\\begin{Shaded}\n\\begin{Highlighting}[]\n\\OtherTok{main ::} \\DataTypeTok{IO}\\NormalTok{ ()}\n\\end{Highlighting}\n\\end{Shaded}\n\n\\end{codelisting}\n\nlst.~\\ref{lst:some_codeblock1}"
+            `test` "\\begin{codelisting}\n\n\\caption{A code block with under\\_score}\\label{lst:some_codeblock1}\n\n\\begin{Shaded}\n\\begin{Highlighting}[]\n\\OtherTok{main ::} \\DataTypeTok{IO}\\NormalTok{ ()}\n\\end{Highlighting}\n\\end{Shaded}\n\n\\end{codelisting}\n\nlst.~\\ref{lst:some_codeblock1}"
           let test1 = test' $ setMeta "codeBlockCaptions" True nullMeta
               infixr 5 `test1`
           codeBlockForTable "some_codeblock1" <> paraText ": A code block"
             <> para (citeGen "lst:some_codeblock" [1])
-            `test1` "\\begin{codelisting}\n\n\\caption{A code block}\n\n\\hypertarget{lst:some_codeblock1}{%\n\\label{lst:some_codeblock1}}%\n\\begin{Shaded}\n\\begin{Highlighting}[]\n\\OtherTok{main ::} \\DataTypeTok{IO}\\NormalTok{ ()}\n\\end{Highlighting}\n\\end{Shaded}\n\n\\end{codelisting}\n\nlst.~\\ref{lst:some_codeblock1}"
+            `test1` "\\begin{codelisting}\n\n\\caption{A code block}\\label{lst:some_codeblock1}\n\n\\begin{Shaded}\n\\begin{Highlighting}[]\n\\OtherTok{main ::} \\DataTypeTok{IO}\\NormalTok{ ()}\n\\end{Highlighting}\n\\end{Shaded}\n\n\\end{codelisting}\n\nlst.~\\ref{lst:some_codeblock1}"
 
 citeGen :: T.Text -> [Int] -> Inlines
 citeGen p l = cite (mconcat $ map (cit . (p<>) . T.pack . show) l) $ text $
@@ -340,42 +373,49 @@ refGen' :: T.Text -> [Int] -> [(Int, Int)] -> M.Map T.Text RefRec
 refGen' p l1 l2 = M.fromList $ mconcat $ zipWith refRec''' (((uncapitalizeFirst p<>) . T.pack . show) `map` l1) l2
 
 refRec' :: T.Text -> Int -> T.Text -> [(T.Text, RefRec)]
-refRec' ref i tit = [(ref, RefRec{refIndex=[(i,Nothing)],refTitle=toList $ text tit,refSubfigure=Nothing})]
+refRec' ref i tit = [(ref, RefRec{refLabel=Just ref,refIndex=[(i,Nothing)],refGlobal=0,refHideFromList=False,refTitle=toList $ text tit,refSubfigure=Nothing})]
 
 refRec'' :: T.Text -> Int -> [(T.Text, RefRec)]
 refRec'' ref i = refRec' ref i ""
 
 refRec''' :: T.Text -> (Int, Int) -> [(T.Text, RefRec)]
-refRec''' ref (c,i) = [(ref, RefRec{refIndex=[(c,Nothing), (i,Nothing)],refTitle=toList $ text "",refSubfigure=Nothing})]
+refRec''' ref (c,i) = [(ref, RefRec{refLabel=Just ref,refIndex=[(c,Nothing), (i,Nothing)],refGlobal=0,refHideFromList=False,refTitle=toList $ text "",refSubfigure=Nothing})]
 
-testRefs' :: T.Text -> [Int] -> [Int] -> Lens' References (M.Map T.Text RefRec) -> T.Text -> Expectation
-testRefs' p l1 l2 prop res = testRefs (para $ citeGen p l1) (set prop (refGen p l1 l2) def) (para $ text res)
+testRefs' :: T.Text -> [Int] -> [Int] -> Prefix -> T.Text -> Expectation
+testRefs' p l1 l2 prop res = testRefs (para $ citeGen p l1) (set (refsAt prop) (refGen p l1 l2) def) (para $ text res)
 
-testRefs'' :: T.Text -> [Int] -> [(Int, Int)] -> Lens' References (M.Map T.Text RefRec) -> T.Text -> Expectation
-testRefs'' p l1 l2 prop res = testRefs (para $ citeGen p l1) (set prop (refGen' p l1 l2) def) (para $ text res)
+testRefs'' :: T.Text -> [Int] -> [(Int, Int)] -> Prefix -> T.Text -> Expectation
+testRefs'' p l1 l2 prop res = testRefs (para $ citeGen p l1) (set (refsAt prop) (refGen' p l1 l2) def) (para $ text res)
 
 testAll :: (Eq a, Data a, Show a) => Many a -> (Many a, References) -> Expectation
-testAll = testState f def
-  where f = References.Blocks.replaceAll defaultOptions
+testAll = testState References.Blocks.replaceAll def
 
-testState :: (Eq s, Eq a1, Show s, Show a1, Df.Default s) =>
-               ([a] -> State s [a1]) -> s -> Many a -> (Many a1, s) -> Expectation
-testState f init' arg res = runState (f $ toList arg) init' `shouldBe` first toList res
+testState :: (Show a1, Eq a1) => ([a2] -> WS [a1]) -> References -> Many a2 -> (Many a1, References) -> Expectation
+testState f init' arg res = runWSWithOptsInit defaultOptions init' (f $ toList arg) `shouldBe` first toList res
+
+runWSWithOptsInit :: Options -> References -> WS a -> (a, References)
+runWSWithOptsInit opts st = flip runState st . flip runReaderT opts . runWS
+
+runWSWithOpts :: Options -> WS a -> (a, References)
+runWSWithOpts opts = runWSWithOptsInit opts def
 
 testRefs :: Blocks -> References -> Blocks -> Expectation
-testRefs bs st rbs = testState (bottomUpM (References.Refs.replaceRefs defaultOptions)) st bs (rbs, st)
+testRefs bs st rbs = testState (bottomUpM References.Refs.replaceRefs) st bs (rbs, st)
 
 testCBCaptions :: Blocks -> Blocks -> Expectation
-testCBCaptions bs res = runState (bottomUpM (Util.CodeBlockCaptions.mkCodeBlockCaptions defaultOptions{Text.Pandoc.CrossRef.Util.Options.codeBlockCaptions=True}) (toList bs)) def `shouldBe` (toList res,def)
+testCBCaptions bs res = runWSWithOpts defaultOptions{Text.Pandoc.CrossRef.Util.Options.codeBlockCaptions=True}
+  (bottomUpM (Util.CodeBlockCaptions.mkCodeBlockCaptions) (toList bs)) `shouldBe` (toList res,def)
 
 testList :: Blocks -> References -> Blocks -> Expectation
-testList bs st res = runState (bottomUpM (References.List.listOf defaultOptions) (toList bs)) st `shouldBe` (toList res,st)
+testList bs st res = runWSWithOptsInit defaultOptions st (bottomUpM References.List.listOf (toList bs))
+   `shouldBe` (toList res,st)
 
-figure :: T.Text -> T.Text -> T.Text -> T.Text -> Blocks
-figure = (((para .) .) .) . figure' "fig:"
+figure :: T.Text -> T.Text -> T.Text -> Maybe T.Text -> T.Text -> Blocks
+figure src title cap malt ref = B.figureWith ("fig:" <> ref, [], [])
+  (caption Nothing $ para $ text cap) $ plain $ figure' src title (fromMaybe cap malt) ""
 
-figure' :: T.Text -> T.Text -> T.Text -> T.Text -> T.Text -> Inlines
-figure' p src title alt ref = imageWith ("fig:" <> ref, [], []) src (p <> title) (text alt)
+figure' :: T.Text -> T.Text -> T.Text -> T.Text -> Inlines
+figure' src title alt ref = imageWith (if T.null ref then mempty else "fig:" <> ref, [], []) src title (text alt)
 
 section :: T.Text -> Int -> T.Text -> Blocks
 section text' level label = headerWith ("sec:" <> label,[],[]) level (text text')
@@ -435,5 +475,12 @@ cit :: T.Text -> [Citation]
 cit r = [defCit{citationId=r}]
 
 infixr 0 =:
-(=:) :: Df.Default r => Lens' r a -> a -> r
-a =: b = a .~ b $ def
+(=:) :: Prefix -> M.Map T.Text RefRec -> References
+a =: b = def
+  & ctrsAt a .~ (refIndex $ last $ M.elems b)
+  & refsAt a .~ b
+  & stGlob %~ (+ fromIntegral (M.size b))
+  & stHiddenHeaderLevel .~
+    if a == PfxSec
+    then [HiddenHeader { hhLevel = 1, hhHidden = False }]
+    else []
